/***************************************************************/
/*                         eml.c                               */
/*　分類知識を機械学習するプログラム　　                       */
/*　　　入力は標準入力から与え，出力は標準出力に出力します     */
/* 使い方　                                                    */
/*       $./eml 　< (入力ファイル名）> (出力ファイル名)        */
/*　入力ファイルでは，属性とカテゴリを0/1で記述します          */
/*　出力ファイルには，学習結果を表示します                     */
/***************************************************************/

/* ヘッダファイルのインクルード*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*記号定数の定義*/
#define TEXTLENGTH 256 /*テキストの最大長*/
#define DATANO 256   /*学習データセットの最大個数*/
#define ATTNO 4 /*属性値の個数*/
#define TERMNO 3 /*論理式の論理和項の個数*/
#define YES 1   /*yesに対応する整数値*/
#define NO 0    /*noに対応する整数値*/
#define SEED 65535    /*乱数のシード*/
#define REPNO 1000    /*分類知識の生成回数*/
#define THVAL 0.7     /*結果出力の下限値*/

/*関数のプロトタイプの宣言*/
int readdata(int dataset[DATANO][ATTNO+1]) ;
                     /*学習データ読み込み*/
int rnd() ; /*    0/1の乱数の生成   */
void genformula(int formula[TERMNO][ATTNO*2]) ;
                                /*論理式の生成*/
double evalformula(int formula[TERMNO][ATTNO*2]
  ,int dataset[DATANO][ATTNO+1],int no) ;/*論理式の評価*/
void purformula(int formula[TERMNO][ATTNO*2]) ; /*論理式の出力*/
int evals(int formula[TERMNO][ATTNO*2]
      ,int subset[ATTNO+1]); /*特定の事例に対する評価*/
int notval(int v) ;/*真理値の反転*/

/*******************/
/*    main()関数   */
/*******************/
int main(int argc,char *argv[])
{
 int no ; /*学習データセットの個数*/
 int i ; /*繰り返しの制御変数*/
 int dataset[DATANO][ATTNO+1]={0} ;
       /*学習データセットの属性とカテゴリを格納*/
 int formula[TERMNO][ATTNO*2]={0} ;
       /*獲得対象となる論理式*/
 double score=0 ;/*論理式の評価値*/

 /*乱数の初期化*/
 srand(SEED) ;

 /*学習データセットの読み込み*/
 no=readdata(dataset) ;

 /*論理式の生成と検査*/
 for(i=0;i<REPNO;++i){
  genformula(formula) ;/*論理式の生成*/
  score=evalformula(formula,dataset,no) ;/*論理式の評価*/
  /*論理式の出力*/
  if(score>THVAL){
   printf("\nスコア：%lf\n",score) ;
   purformula(formula) ;
  }
 }
 
 return 0 ;
}

/************************/
/*   readdata()関数     */
/* 学習データの読み込み */
/************************/
int readdata(int dataset[DATANO][ATTNO+1])
{
 int no=0 ; /*学習データセットの個数*/
 int att=0 ; /*学習データセットの属性*/

 /*データの入力*/
 while(scanf("%d",&dataset[no][att])!=EOF){
  ++ att ;
  if(att>ATTNO){/*次のデータ*/
   att=0 ;
   ++no ;
  }
 }
 return no ;
}


/************************/
/*     rnd()関数        */
/*    0/1の乱数の生成   */
/************************/
int rnd()
{
 int rndno ;/*生成した乱数*/

 while((rndno=((double)rand()/RAND_MAX)*2)==2) ;

 return rndno;
}

/************************/
/*   genformula()関数   */
/*    論理式の生成      */
/************************/
void genformula(int formula[TERMNO][ATTNO*2])
{
 int i,j ;/*ループ制御の変数*/

 for(i=0;i<TERMNO;++i)
  for(j=0;j<ATTNO*2;++j)
   /*0又は1の乱数によって値を決定*/
   formula[i][j]=rnd() ;
}

/************************/
/*   putformula()関数   */
/*    論理式の出力      */
/************************/
void purformula(int formula[TERMNO][ATTNO*2]) 
{
 int i,j ;/*ループ制御の変数*/

 for(i=0;i<TERMNO;++i){
  for(j=0;j<ATTNO*2;++j)
   printf("%d ",formula[i][j]) ;
  printf("\n") ;
 }
}
 

/************************/
/*   evalformula()関数  */
/*    論理式の評価      */
/************************/
double evalformula(int formula[TERMNO][ATTNO*2]
  ,int dataset[DATANO][ATTNO+1],int no)
{
 int i ;/*ループ制御の変数*/
 int point=0 ;/*正解の個数*/

 /*学習セットの各事例に論理式を適用*/
 for(i=0;i<no;++i){
  point+=evals(formula,dataset[i]) ; 
 }

 return (double)point/no ;/*評価値を戻す*/
}

/************************/
/*    evals()関数       */
/*特定の事例に対する評価*/
/************************/
int evals(int formula[TERMNO][ATTNO*2]
  ,int subset[ATTNO+1])
{
 int orval,andval,result ;/*計算の途中経過値*/
 int i,j ;/*ループ制御の変数*/

 andval=YES ; 
 /*論理積で項を結んだ，式全体の評価*/
 for(i=0;i<TERMNO;++i){
  orval=NO ;
  /*論理和で結んだ項の評価*/
  for(j=0;j<ATTNO*2;++j){
   if((j%2)==0) orval+=subset[j/2]*formula[i][j] ;
   else orval+=notval(subset[j/2])*formula[i][j] ;
  }
  if(orval==NO) andval=NO ;
  else andval*=YES ; 
 }

 if(andval==subset[ATTNO]) return 1 ;/*結果が一致*/
 else return 0 ;/*不一致*/
}


/************************/
/*   notval()関数       */
/*   真理値の反転       */
/************************/
int notval(int v)
{
 if(v==YES) return NO ;
 else return YES ;
}

